/**
 * Lecture notes:
 * 1) Optimize this pice of code (hint: there are at least two performance-bugs)
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <CL/cl.h>

#include "../misc/ocl_util.h"


#define BLOCKSIZE 32


/// y <- alpha * x + y (CPU VERSION)
void saxpy_ref(int num_elements, float * const x, float alpha, float* y);


/// compares the results of the GPU version against CPU
void check_result(int num_elements, float * const x, float alpha, float * const y, float *y_copy);



int main(int argc, char** argv)
{
   int i;
   int num_elements = 10240;

   size_t size = sizeof(float) * num_elements;
   printf("Allocates %f MB\n", 2.0f * ((float)size)/1024/1024);


   //allocate host memory
   float *x_h, *y_h, *y_copy;
   x_h = (float*) malloc(size);
   y_h = (float*) malloc(size);
   y_copy = (float*) malloc(size);


   //init x and y with random numbers in [0,1]
   for(i=0;i < num_elements; ++i){
      x_h[i] = ((float)rand())/RAND_MAX;
      y_h[i] = ((float)rand())/RAND_MAX;
      y_copy[i] = y_h[i];
   }

   //initialize opencl context;
   cl_int error;
   ocl_context_t ocl_context;
   ocl_context_init(&ocl_context, "saxpy.cl");


   //allocate device memory
   cl_mem x_d, y_d;
   x_d = clCreateBuffer(ocl_context.context, CL_MEM_READ_ONLY, size, NULL, &error);
   HANDLE_ERROR( error );
   y_d = clCreateBuffer(ocl_context.context, CL_MEM_READ_WRITE, size, NULL, &error);
   HANDLE_ERROR( error );


   //copy data from host to device
   HANDLE_ERROR( clEnqueueWriteBuffer( ocl_context.command_queue[0], 
                                       x_d, CL_TRUE, 0, size, 
                                       (void*) x_h, 0, NULL, NULL) ); //CL_TRUE, indicates blocking!
   HANDLE_ERROR( clEnqueueWriteBuffer( ocl_context.command_queue[0], 
                                       y_d, CL_TRUE, 0, size, 
                                       (void*) y_h, 0, NULL, NULL) );

   //Set the arguments of the saxpy kernel
   float alpha = 2.0;
   HANDLE_ERROR( clSetKernelArg (ocl_context.saxpy, 0, sizeof(cl_int), (void*)&num_elements) );
   HANDLE_ERROR( clSetKernelArg (ocl_context.saxpy, 1, sizeof(cl_mem), (void*)&x_d) );
   HANDLE_ERROR( clSetKernelArg (ocl_context.saxpy, 2, sizeof(cl_float), (void*)&alpha) );
   HANDLE_ERROR( clSetKernelArg (ocl_context.saxpy, 3, sizeof(cl_mem), (void*)&y_d) );

   //launch kernel
   const size_t local_work_size = BLOCKSIZE;
   const size_t num_groups = (num_elements + BLOCKSIZE - 1)/ BLOCKSIZE;
   const size_t global_work_size = num_groups * BLOCKSIZE;
   HANDLE_ERROR(  clEnqueueNDRangeKernel (ocl_context.command_queue[0],
                  ocl_context.saxpy, 1, NULL,
                  &global_work_size, &local_work_size,
                  0, NULL, NULL) ); 


   //copy data from device to host
   HANDLE_ERROR( clEnqueueReadBuffer( ocl_context.command_queue[0], 
                                       y_d, CL_TRUE, 0, size, 
                                       (void*) y_h, 0, NULL, NULL) );


   //verify the result
   check_result(num_elements, x_h, alpha, y_h, y_copy);

      
   //Cleanup
   ocl_context_free(&ocl_context);
   clReleaseMemObject(x_d); clReleaseMemObject(y_d);

   free(x_h); free(y_h); free(y_copy);

   return 0;
}


void saxpy_ref(int num_elements, float * const  x, float alpha, float* y)
{
   int i;
   for(i=0;i < num_elements; ++i)
      y[i] += alpha * x[i];

}


/// compares the cpu and gpu results element-wise and checks if the maximal relative error is larger than some threshold
void check_result(int num_elements, float * const  x, float alpha, float * const y, float* y_copy)
{
   //start cpu calculation
   saxpy_ref(num_elements, x, alpha, y_copy);

   float max_error = 0.0f;
   int i;
   for(i=0;i < num_elements; ++i)
      if(fabs((y[i] - y_copy[i]) / y[i]) > max_error)
         max_error = fabs((y[i] - y_copy[i]) / y[i]);

   if(max_error < 1e-5)
      printf_green((char*)"Result OKAY!\n");
   else
      printf_red((char*)"ERROR: Result-check FAILED!\n");
}


